Title:  Ubuntu Core Product Integration Agent Reference Guide
Version: 20151124

[TOC]

Installation
============

Follow the steps outlined in the [Tutorial](tutorial.html#lab-setup).

Configuration
==============

The agent’s configuration options:

<table class="table">
<tr>
    <th>Key         <th>Description
<tr>
    <td>id            <td>A unique identifier for each agent instance, e.g. 'lab01-agent007'. This is useful, for example, to know which agent executed a specific test run.
<tr>
    <td>server        <td>URL for the Product Integration server
<tr>
    <td>interval      <td>Polling interval for retrieving new Tests
<tr>
    <td>provcommand   <td>See below under [Provisioning Kit Scripts](#provisioning-kit-scripts)
<tr>
    <td>setupcommand  <td>See below under [Provisioning Kit Scripts](#provisioning-kit-scripts)
<tr>
    <td>testcommand   <td>See below under [Provisioning Kit Scripts](#provisioning-kit-scripts)
<tr>
    <td>execution_basedir            <td>Path under which Provisioning Kit scripts will be run
<tr>
    <td>logging_basedir            <td>Path for storing logs
<tr>
    <td>network_timeout            <td>Timeout for network operations, defaults to 10 seconds.
<tr>
    <td>consumer_key            <td>oauth identifier
<tr>
    <td>consumer_secret            <td>oauth keys and secrets
<tr>
    <td>token_key            <td>oauth keys and secrets
<tr>
    <td>token_secret            <td>oauth keys and secrets
</table>

The agent also supports a ```[environment]``` section in the configuration
file. The contents of this section will be passed, unmodified, to the
provisioning kit’s scripts environment.

Troubleshooting
================

Agents that cannot register will show error messages in their terminal and in the logfile. For example, if there is a network failure,
the agent will show an error like this, and keep retrying:

    requests.exceptions.ConnectionError: ('Connection aborted.', ConnectionRefusedError(111, 'Connection refused'))


Test Opportunity Data Reference
=======================


From our system, the agent will obtain "Test Opportunities". A test opportunity is a JSON object describing the tests
to run, and the information from where that test opportunity was generated. The important fields of note are as follows:

<table class="table">
<tr>
    <th>Key     <th>Description
<tr>
    <td>image_reference        <td>A JSON value from either the Manifest or submitted image (image-based workflow).
        <p>The provcommand should use this imformation to provision the image onto the test device.
<tr>
    <td>id        <td>A unique ID for this test opportunity
<tr>
    <td>image_unique_id        <td>As with image_reference, from either the Manifest or submitted image. Should be used by the provcommand to verify the integrity of the image_reference.
<tr>
    <td>test_payload        <td>An JSON value that the setupcommand and testcommand can use to setup and run tests.
    <p>E.g. this could be the JSON object:
    <pre>
  {
    "test_url": "http://example.com/foo.tgz",
    "test_cmd": "run_tests.sh"
  }
    </pre>
    <p>or just a string like "http://example.com/foo.tgz", etc.
    <p>The important thing is that those two commands will understand what to do with that data. The field is optional on the server in which case there’s a null in it.
<tr>
    <td>extra_snaps        <td>A JSON array of (snap_name, revision)
    tuples. These are generated from a combination of Manifest Snaps
    and Snap revision information advertised to the system. They should be used by the testsetup command to install or upgrade the listed Snaps.
<tr>
    <td>spec_name        <td>The name of the test specification that generated this test opportunity
<tr>
    <td>test_spec        <td>A copy of the test specification that generated this test opportunity
<tr>
    <td>manifest (optional)        <td>A copy of a manifest related to this test opportunity
<tr>
    <td>image (optional)        <td>A copy of the image information that triggered this test opportunity
</table>


An example Test Opportunity is:

    {
    'created_at': '2015-08-04T16:22:23.350261',
    'extra_snaps': [],
    'id': '117a9a67-f066-4afa-ae05-d039b4cee457',
    'image': {'image_name': 'docs-img',
            'image_reference': 'http://foo.com/bar',
            'image_unique_id': 'DEADBEEF_8870'},
    'image_name': 'docs-img',
    'image_reference': 'http://foo.com/bar',
    'image_unique_id': 'DEADBEEF_8870',
    'platform': 'amd64',
    'spec_name': 'docs-spec_8870',
    'test_payload': None,
    'test_spec': {'created\_at': '2015-08-04T16:22:15.160433',
                'id': '12ee7abd-a11f-4a95-b7e2-6c54d9f3780a',
                'image_name': 'docs-img',
                'name': 'docs-spec_8870',
                'platform': 'amd64',
                'test_payload': None}}


Provisioning Kit Scripts
================

The provisioning kit consists of 3 binaries, with paths defined in
the agent's config. For example: "/opt/prov/pk-provision",
"/opt/prov/pk-setup", and "/opt/prov/pk-test".

The agent will store the test opportunity in a file in the working
directory and execute each command in order, passing the path to each
one, like this:

-   pk-provision spi\_test\_opportunity.json
-   pk-setup spi\_test\_opportunity.json
-   pk-test spi\_test\_opportunity.json

The CWD for the commands will be a per-test-opportunity
directory.

**The test result will be stored by the provisioning kit in $CWD/spi_test_result.json and will be uploaded to the server by the agent.**



The content of spi\_test\_result.json should be a valid JSON
object, with a maximum sizze of 4000 bytes, and will be
shown in the Test Result output. For example, if the test result is
this:

    {"foo": "bar"}

Then the Test Result will contain that data under "result\_payload":

    {
        "image_reference": "http://foo.com/bar",
        "test_status": "PASSED",
        "result_payload": {
            "foo": "bar"
        },
        "image_unique_id": "DEADBEEF_1845",
        "platform": "amd64",
        "agent_id": "docs-id",
        "id": "a03cb2af-8bcb-4147-b303-829d61de6d4a"
    }



These binaries will probably be thin wrappers around the
provisioning and testing tools the customer already uses, passing
information from the test opportunity to those tools. Any data that
needs to be passed from a stage to the next, needs to be handled by the
provisioning kit binaries. For example:



-   pk-provision setups a device, obtains a IP address
-   pk-setup needs that IP address



To achieve that, pk-provision will store that data in a well-known
location in some manner.



The pk-provision binary is responsible for provisioning the device
where the tests will run. For example:



-   Reset a beaglebone board
-   Retrieve the image referenced by image\_reference
-   Maybe verify it using image\_unique\_id if it were a hash of
    the image.
-   Flash the device
-   Boot the device
-   Obtain the IP address and store it in a well known
    location
-   Evaluate success/failure
-   Exit with appropriate success/failure value



The pk-setup binary is responsible for setting up any required
data or application. For example:



-   Connect to the device using the IP address obtained in
    provisioning phase
-   Install packC version 3, upgrade packD to version 5
-   Evaluate success/failure
-   Exit with appropriate success/failure value



The pk-test binary is responsible for obtaining and running the
tests. For example:



-   Retrieve the testcase using the information in
    test\_payload
-   Copy data and test code to the device using IP address
    obtained in provisioning phase
-   Execute tests, logging output, save artifacts
-   Evaluate success/failure
-   Store result in spi\_test\_result.json
-   Save valuable test artifacts and logs to proper
    locations
-   Exit with appropriate success/failure value



After the test run ends, the agent will feed the results back into
the system for the user to read and evaluate.



It's reasonable that in some lab settings the operating procedure
will not match this description, and the lab technician will collapse
steps, or add extra steps, for example, by making setup part of
provisioning, or even making all the steps a single process. The agent
is supposed to be flexible enough to handle that.

Log and Artifact Management
====================


Running tests in the lab devices will generate a number of files
which will be of interest to developers and QA engineers to diagnose and
solve problems. The Product Integration system is not intended as a repository for those
files  and artifacts.



It's best practice to upload those files to a well established,
permanent location, like SWIFT, S3, Dropbox, etc, and use the result
payload to store references to them. That way, users can easily access
them when needed.

Test Results
=========

There are a number of actors involved in running a test, and there
are different ways in which a test can fail. The separation in three
binaries aims to model this to some extent.

For example, we could have the following actors:

-   A lab technician, setting up agents and devices
-   A QA engineer setting up tests
-   A developer releasing new Snap or Image versions

Each provisioning kit binary will either succeed (exit 0) or fail
(exit 1).

-   If pk-provision fails, the test result will be a
    PROVFAIL
-   If pk-setup fails, the test result will be a SETUPFAIL
-   if pk-test fails, the test result will be TESTFAIL
-   if pk-test succeeds, the test result will be SUCCESS

A PROVFAIL would usually indicate a problem in the lab setup for
the lab technician to check. In some cases it would indicate a problem
in the base Image.

A SETUPFAIL could indicate a problem in the base Image, or in the
developer's apps.

A TESTFAIL indicates a problem in the developer's application or
in the QA person's testcase.

Test Retries
============

Test Opportunities consumed by an Agent that do not result in uploaded Test Results (```PASS``` or ```FAILED```) within a period of
**6 hours** will be re-delivered, i.e. will be available to be consumed again in the same context 
(same agent or another agent with access to the same products).
The server understands that the Test Opportunity in question has met a malfunctioning agent or got lost
due to a network problem, thus have to be retried. Tests genuinely taking longer than this arbitrary period will be dispatched continuously
until we can allow users to specify effective timeouts according to
their Test Specs.

Agent Upgrades
==============

Upgrading the Agent is accomplished via the following process:

- Stop the running Agent
- Move the Agent directory out of the way
- Download and install the latest agent (see [Tutorial](tutorial.html#lab-setup))
- Copy the Agent's config.ini from the old directory to the new one

