#!/usr/bin/env python

import argparse
import sys

import markdown

TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{title[0]}</title>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/anchor-js/1.3.0/anchor.min.js"></script>
    <link href="https://maxcdn.bootstrapcdn.com/bootswatch/3.3.5/united/bootstrap.min.css" rel="stylesheet">
    <style>
      h2{{margin:40px 0 20px;border-bottom:1px solid #eee}}
      .docs-footer{{margin:5em 0}}
    </style>
</head>
<body>
<div class="container">
  <div class="page-header">
    <h1>{title[0]}</h1>
    <small class="text-muted">Version: {version[0]}</small>
  </div>
  {body}
</div>
<footer class="docs-footer"><p class="text-center"><a href="#top">Back to top</a></p></footer>

<script type="text/javascript">
anchors.options = {{
  placement: 'left'
}};
anchors.add('h2,h3');
</script>

</body>
</html>
"""


def parse_args(args=None):
    d = 'Make a complete, styled HTML document from a Markdown file.'
    parser = argparse.ArgumentParser(description=d)
    parser.add_argument('mdfile', type=argparse.FileType('r'), nargs='?',
                        default=sys.stdin,
                        help='File to convert. Defaults to stdin.')
    parser.add_argument('-o', '--out', type=argparse.FileType('w'),
                        default=sys.stdout,
                        help='Output file name. Defaults to stdout.')
    return parser.parse_args(args)


def main(args=None):
    args = parse_args(args)
    text = args.mdfile.read()
    extensions = ['extra', 'meta', 'toc']
    md = markdown.Markdown(extensions=extensions, output_format='html5')
    html = md.convert(text)
    doc = TEMPLATE.format(body=html, toc=md.toc, **md.Meta)
    args.out.write(doc)


if __name__ == '__main__':
    sys.exit(main())
