Title:   Getting Started with Snappy Product Integration
Version: 20151030

Snappy Product Integration (SPI) is a system for automated integration
testing of products built upon Ubuntu Core. Once configured, SPI will
monitor for updates to components of your product and run automated
tests against target devices to determine if those updates are
compatible, and allow you to control the flow of those updates to your
customers.

This tutorial will guide you step-by-step to create a snap for an application,
define a product that includes that application, create a laboratory
environment where integration tests for that product can run, and set up the
SPI system so that it automatically executes tests in that laboratory when new
versions of the snap become available.

The document is organized as follows:

-   **[What you will need:](#what-you-will-need)** A description of everything needed to complete the tutorial.
-   **[Lab setup:](#lab-setup)** Create and configure your own laboratory.
-   **[Product update testing:](#product-update-testing)** Walkthrough and configuration of how to test a product software update and check results.
-   **[Other testing scenarios:](#other-testing-scenarios)** Triggering tests
    when the operating system snap or your Gold Master image changes.

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Conventions</h3>
  </div>
  <div class="panel-body">
    From time to time you will need to execute commands in a terminal. This
    will be indicated using <code>this particular font and background color</code>,
    normally on a separate line. This font will also be used to specify
    filenames or any other cases where spelling matters.
  </div>
</div>

## What you will need

In order to complete this tutorial, you will need the following resources:

-   A host computer running [Ubuntu 14.04 (64-bit)](http://www.ubuntu.com/download/desktop), with an SD card reader and at least 6 GB free drive space.
-   A [Raspberry Pi 2 Model B](https://www.raspberrypi.org/products/raspberry-pi-2-model-b/).
-   A 4GB (minimum) Micro SDHC card.
-   The host computer and Raspberry Pi connected on the same local ethernet network.
-   Access to the Internet from the host computer.
-   An optional serial USB cable, for debugging the RPI boot sequence.
-   OAuth credentials tied to an SPI Organization. This will be covered below
    during [Lab setup](#lab-setup).

## Lab setup

In order to run tests on your product, SPI needs a host computer to serve as a
bridge between your device and itself. This computer can be as simple as your
desktop, or part of a full device laboratory.

For the purposes of this tutorial, your desktop computer will serve as the
host and you will use a Raspberry Pi as the product device. The host computer
will run a small application called the SPI Agent, which will communicate with
Canonical's servers and manage integration tests on the Raspberry Pi.

### Setting up the host computer

You will first need to install and configure the SPI Agent. The most recent
release of the Agent can always be found at [this location](https://spi.canonical.com/assets/agent.tar.gz).

Download and install the Agent from your host computer as follows:

    wget https://spi.canonical.com/assets/agent.tar.gz
    tar -zxf agent.tar.gz
    cd agent-binaries

Also install the software dependencies of the Agent:

    sudo apt-get install python-requests pv


### Configuring the Agent

Once the Agent is installed, you will need to create its configuration:

    ./agent.py --default-config > config.ini

The Agent authenticates using Ubuntu One. To use it, first create an
[Ubuntu One account](https://login.ubuntu.com), then instruct the Agent to put
your credentials in a configuration file:

    ./agent.py --interactive-login config.ini

<div class="panel panel-danger">
  <div class="panel-heading">
    <h3 class="panel-title">Security warning</h3>
  </div>
  <div class="panel-body">
    The config file now has data (<code>consumer_secret</code>,
    <code>token_secret</code>, and <code>token_key</code>) that are equivalent
    to your password. You are responsible for storing this securely. The device
    on which the Agent runs should be protected from unauthorized access.
  </div>
</div>


### Setting up your organization

All users of SPI are each part of an Organization. Your Agents need to be
registered using a named reference of your physical organization, an Organization ID.
Organization IDs can only contain lowercase ASCII letters, numbers and hyphens (-). Hyphens can not occur at the start or end of the chosen value.
Maximum length for an Organization ID is 20 characters.
For example, if your company name is "Widgets R' Us", you could create an organization ID of "widgets-r-us"

To set up or add users to an Organization:

1.  In the Agent configuration created when you logged in, there is a key of
the form ```consumer_key = xYzZY```.
2.  Contact your Canonical representative. Provide them with your chosen
Organization ID and this aforementioned ```consumer_key```. Once your
representative has added this key into the Organization, any agent configured
with this consumer key will now have access to your Organization's data.

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Consumer key safety</h3>
  </div>
  <div class="panel-body">
     The consumer key is equivalent to your user ID, not password, and is safe
     to send in email.
  </div>
</div>

### Connecting the host computer to your organization

Edit the ```config.ini``` and replace the value of the ```organization_id``` field in the ```config.ini``` file with your Organization ID, and ```platform``` to "spi-tutorial-raspi2" for the sake of this tutorial.

Later, we will need to use the Organization ID in the command line examples.
You can set up an environment variable to save on typing:

    export ORGANIZATION="Your Organization ID"

Now you can verify your credentials are properly associated with your Organization by registering your Agent in the system:

    ./agent.py --register config.ini

You should see a message similiar to the following:

    [agent-queue] Registered at https://spi.canonical.com/orgs/my-org/labs/agent-groups/default

Please double check your credentials and config.ini to ensure the
```organization_id``` is correct, if you instead see an error similiar to the
following:

    {
      "id": "unauthorized",
      "message": "unauthorized for organization"
    }



Once successfully registered, you can do a test run of the Agent:

    ./agent.py -v --once config.ini

If successful, this will produce output similar to the following:

    [agent-queue] Registered at https://spi.canonical.com/orgs/my-org/labs/agent-groups/default
    [agent-queue] No tests at https://spi.canonical.com/orgs/my-org/tests/pending/spi-tutorial-raspi2/default/agent-queue

## Product update testing

Now it is time to create your first product based on Ubuntu Core. On your host computer, first install the latest Snappy tools:

    sudo add-apt-repository ppa:snappy-dev/tools
    sudo apt-get update
    sudo apt-get dist-upgrade
    sudo apt-get install snappy-tools

See [Snappy Getting Started Guide](https://developer.ubuntu.com/en/snappy/start/) for further details.

### Creating a product image

Next, you will create an image for RPI2 from the rolling/edge channel, with our test-snap pre-installed:

    wget https://spi.canonical.com/assets/test-snap_1_all.snap
    sudo ubuntu-device-flash core rolling --channel edge --oem pi2.canonical --device raspi2_armhf \
        --developer-mode --install test-snap_1_all.snap --install webdm -o spi-test_pi2.img

Alternatively, you can download an already prepared sample image:

    wget https://spi.canonical.com/assets/spi-test_pi2.img.xz

If you elect to use the sample image, take care to verify that the MD5 hash for
the downloaded file matches the value shown below:

    md5sum spi-test_pi2.img.xz
    be7cf02999e576aa5eb76e7e85713812  spi-test_pi2.img.xz

If using the sample image, you will also need to decompress it before
continuing:

    unxz spi-test_pi2.img.xz

<div class="panel panel-info">
  <div class="panel-heading">
    <h3 class="panel-title">Modelling your product</h3>
  </div>
  <div class="panel-body">
     When working with a real product, you will substitute in your custom OEM
     and application snaps. See the <a
     href="https://developer.ubuntu.com/en/snappy/guides/oem/">Snappy OEM
     documentation</a> for more information.
  </div>
</div>

### Provisioning the Raspberry Pi

Because remotely flashing the Raspberry Pi is a feature still in active
development, you will need to manually flash an SD card with the image.

On the host machine, insert your SD card and determine which device it is
(using ```df -h``` or other tools), then copy the image to your flash device:

    cat spi-test_pi2.img | pv -s 4g | sudo dd of=/dev/sdX bs=32M
    sync

<div class="panel panel-danger">
  <div class="panel-heading">
    <h3 class="panel-title">Careful device selection</h3>
  </div>
  <div class="panel-body">
     Be careful select the correct output device ('of' flag). The dd command
     will offer no prompt, nor any protection from accidentally selecting your
     hard drive.
  </div>
</div>

This should take less than five minutes on most modern hardware. For additional instructions see [Installing Images](https://www.raspberrypi.org/documentation/installation/installing-images/).

When the flashing is complete, insert the SD card in your Raspberry Pi 2, power
it on, and wait for it to boot. This typically takes less than a minute to
complete.

You can check whether it's online just by running:

    ping webdm.local

If the device is not accessible, you can troubleshoot by using a serial cable
to [obtain a console on the Raspberry Pi](http://elinux.org/RPi_Serial_Connection).
Note that the pins are the same for both versions of the Raspberry Pi.

Alternatively, it may be that the image is not properly written. If the device
fails to boot, put the SD card back in the host machine and write the image a
second time. Verify the write was succesful by confirming that the following
commands both return the same hash:

    sudo dd if=/dev/sdX | head -c `stat -c "%s" spi-test_pi2.img` | pv -s `stat -c "%s" spi-test_pi2.img` | sha1sum
    sha1sum spi-test_pi2.img

### Product and test setup

For SPI to generate tests in case of new snap revisions, some configuration needs to be done.

The Manifest will describe how your product is built, and has the basic information that allows SPI to detect changes that affect your product when a new snap revision is available. When this happens, SPI will use the information in the Test Specification to trigger the corresponding tests.

To create a Manifest, you need to issue an HTTP request specifying the following:

-   ```image_name```: the name of the image you will be testing, to relate the Manifest to the Test Specs we will be creating later
-   ```release```: the Ubuntu Core release this products aims for (currently "15.04-core" or "rolling-core" as known by the Snappy Store)
-   ```base_image_reference```: a pointer to the base image, in this case a URL pointing to the image binary
-   ```snaps```: the list of snaps that that comprise the product

In the agent directory on your computer (where you already authenticated during
the [Lab setup](#lab-setup) above), execute the following:

    ./api_example.py -X POST config.ini https://spi.canonical.com/orgs/$ORGANIZATION/products/manifests --data '{
        "image_name": "tutorial-raspi2-image",
        "release": "rolling-core",
        "base_image_reference": "https://spi.canonical.com/assets/spi-test_pi2.img.xz",
        "snaps": [
            ["ubuntu-core.canonical", null, false],
            ["test-snap.spi-test", 1, false],
            ["webdm.canonical", null, false]]
    }'

Note "test-snap" is our sample application snap that we will be using to trigger the tests when a new version becomes available, and we start here with sequence 1 in our base image.

Next you will define your test scenario via a Test Specification, with the following fields:

-   ```name```: a unique reference to identify tests and their results
-   ```image_name```: the same that we put in the Manifest above
-   ```platform```: defines the target device and is used to route tests to the proper agents.
-   ```test_payload```: information for the test to be run; in this case we put where the tests can be downloaded, and the command to trigger their execution

Execute the following command to create this Test Specification:

    ./api_example.py -X POST config.ini https://spi.canonical.com/orgs/$ORGANIZATION/tests/specs --data '{
        "name": "Tutorial Example Test Specification",
        "image_name": "tutorial-raspi2-image",
        "platform": "spi-tutorial-raspi2",
        "test_payload": {
            "tarball_url": "https://spi.canonical.com/assets/test_tarball.tar.gz",
            "test_cmd": "./test"
        }
    }'

At this point SPI is configured for your product and tests, and will supervise
the Snappy Store for new versions of the snaps listed in your Manifest.

### Triggering your first test run

Normally the system will automatically detect new revisions of snaps as they
are published to the Store, and manual steps will not be required. However, for
the sake of having an update event to work with as part of this tutorial, we
can manually notify the system that a new version of our snap is available:

    ./api_example.py -X POST config.ini https://spi.canonical.com/orgs/$ORGANIZATION/snap-revisions --data '{
        "name": "test-snap.spi-test",
        "revision": 2
    }'

In the response there will be ```test_opportunity_id``` and
```image_unique_id``` fields. Please make note of them, as they will be used in
the next section.

### Agent Provisioning Kit setup

At this point the system has generated a message that is waiting to be consumed
by your Agent. When that message arrives, the Agent will try to run the
commands to provision, set up, and run the tests. Together these scripts are
called a "Provisioning Kit." A sample kit has been provided for this tutorial
and as a starting point for your own projects.

Edit ```config.ini``` and replace the following three fields:

    provcommand = PATH_TO_AGENT/rpi2-sample-provkit/provision
    setupcommand = PATH_TO_AGENT/rpi2-sample-provkit/setup
    testcommand = PATH_TO_AGENT/rpi2-sample-provkit/runtest

<div class="panel panel-warning">
  <div class="panel-heading">
    <h3 class="panel-title">Path sensitivity</h3>
  </div>
  <div class="panel-body">
     The paths in <code>config.ini</code> must be absolute. Be sure to replace
     the correct values.
  </div>
</div>

Now, start the Agent to consume that test opportunity message and start running
the test:

    ./agent.py --once config.ini

As remote flashing of the Raspberry Pi is still in active development, the
Raspberry Pi will need to be set up manually. After receiving the test to run,
the Agent will output instructions explaining how to obtain an image and write
it to the device. Please follow those instructions and let the Agent know when
you are ready by pressing the Enter key when prompted.

After executing the test, the Agent will automatically retrieve the
results and send them back to Canonical's servers. You can query these results
in the next section, [Checking test progress and results](#checking-test-progress-and-results).

Note that during the provisioning you will be instructed to flash the Raspberry
Pi's card again; for the case of the tutorial you may skip this step since you
flashed the image earlier during the "[Provisioning the Raspberry Pi](#test-provisioning-the-raspberry-pi)" step.


### Checking test progress and results

At any time, you can check the progress of tests generated in the
system (replace ```<test_opportunity_id>``` with the one you saved from
"[Triggering your first test run](#triggering-your-first-test-run)" above):

    ./api_example.py config.ini \
    https://spi.canonical.com/orgs/$ORGANIZATION/tests/events?test_opportunity_id=<test_opportunity_id>

This produces a large JSON document containing all the events that relate to this test, for example:

    {"event_logs": [
        {
        "updated_at": "2015-10-28T18:35:38.197000+00:00",
        "events": [
            {
            "timestamp": "2015-10-28T18:34:56.614000+00:00",
            "event_type": "QUEUED_FOR_AGENTS"
            },
            {
            "platform": "spi-tutorial-raspi2",
            "timestamp": "2015-10-28T18:34:58.988000+00:00",
            "agent_group": "default",
            "agent_id": "special",
            "event_type": "PICKED_BY_AGENT"
            },

The ```event_type``` field will describe what happened, there is a timestamp, and extra information related
to the event.

It will also contain all the information relevant to the test, such as:

      "organization_id": "myorg",
      "queued_at": "2015-10-28T18:34:56.614000+00:00",
      "test_spec_id": "2991e1d1-ae89-4afe-9801-140207526f21",
      "spec_name": "Tutorial Example Test Specification",
      "image_unique_id": "5e6ff512-87b6-427e-917d-615c4d1c0aec",

As well as the test itself in the ```test_opportunity``` field.

You can also check the results for an specific image (you can get the image
id from the response of the query when you told the system above that a
new snap revision was available):

    ./api_example.py config.ini \
    https://spi.canonical.com/orgs/$ORGANIZATION/tests/events?image_unique_id=<unique_image_id>

You can see the results produced for all products within an organization by running:

    ./api_example.py config.ini https://spi.canonical.com/orgs/$ORGANIZATION/tests/events

All tests containing 'RESULT_POSTED' events will have results to be retrieved from SPI:

    ./api_example.py config.ini https://spi.canonical.com/orgs/$ORGANIZATION/results

Or you may directly poll for specific results using filters, as below:

    ./api_example.py config.ini https://spi.canonical.com/orgs/$ORGANIZATION/results?image_unique_id=<image_unique_id>
    ./api_example.py config.ini https://spi.canonical.com/orgs/$ORGANIZATION/results?image_name=tutorial-raspi2-image

The result query will return a document describing one or more test results.
Each will contain, among other things, the pass or failure status of the test
run, as well as all related information such as the manifest, test spec, and so
on:

    {
    "test_results": [
        {
        "organization_id": "myorg",
        "created_at": "2015-10-28T18:34:56.614000+00:00",
        "result_id": "7348dac1-9191-454c-93f0-cedb463aa477",
        "test_status": "PASSED",
        "manifest": {
        ...
      "test_spec": {
        "platform": "spi-tutorial-raspi2",
        "image_name": "tutorial-raspi2-image",
        "test_payload": {
          "test_cmd": "./test",
          "tarball_url": "https://spi.canonical.com/assets/test_tarball.tar.gz"
        },
        "name": "Tutorial Example Test Specification",
        "id": "2991e1d1-ae89-4afe-9801-140207526f21",
        "created_at": "2015-10-28T16:42:42.989056",
        "active": true
      },
    ...

### Artifacts and logs

Logs containing the output and errors of the provisioning kit's scripts are always stored in ```logs/```.
Saving those logs and any artifacts the provisioning kit creates is the responsibility of your test command.

The example provisioning kit will create a tarball with the test opportunity, the test result
and all the provisioning kit logs, using this code:

    ARTIFACTS = [
        'spi_test_result.json',
        'spi_test_opportunity.json',
        'logs/provisioning.log',
        'logs/setup.log',
        'logs/testing.log'
    ]

    [...]

    artifact_dir = os.getenv('ARTIFACT_DIR', '/tmp')
    save_cmd = 'tar czvf {} {}'.format(
        os.path.join(artifact_dir, test_opp['id'] + '.tar.gz'),
        ' '.join(ARTIFACTS),
    )
    log.info('Saving artifact with %s', save_cmd)
    subprocess.check_call(save_cmd, shell=True)

Here, the tarball is stored in /tmp by default, but you can override it in the ```[environment]``` section of your ```config.ini```:

    [environment]
    ARTIFACT_DIR='/home/user/artifacts'

## Other testing scenarios

Future revisions of this document will cover operating system update testing and
Gold Master candidate testing.
