#!/usr/bin/python3

import json
import requests
import subprocess
import sys


def write_output(res_dict):
    with open('spi_test_result.json', 'w') as outf:
        json.dump(res_dict, outf)


def ok(res_dict):
    write_output(res_dict)
    print('TEST PASSED: ', res_dict)
    exit(0)


def fail(msg):
    write_output(dict(msg=msg))
    print('TEST FAILED: ', msg)
    raise


def fetch_and_run():
    res = {}
    # read our test opportunity
    test_opp_file = sys.argv[1]
    print('Test Opportunity file: ', test_opp_file)
    with open(test_opp_file, 'r') as inf:
        test_opp = json.loads(inf.read())
        print('Test Opportunity is: ')
        print(json.dumps(test_opp, indent=2))

    print(test_opp['test_payload'])
    try:
        test_payload = test_opp['test_payload']
    except:
        fail('test_payload is not json: {}'.format(test_opp['test_payload']))

    # our tutorial Provisioning Kit (this script)
    # expects the following in test_payload:
    #   { "tests_url": <some url>, "tests_cmd": <some cmd> }
    if 'tests_url' not in test_payload or 'tests_cmd' not in test_payload:
        fail('bad test_payload: {}'.format(test_payload))

    # 'download' our test content
    # this could be a tarball that we save and untar
    resp = requests.get(test_payload['tests_url'])
    if resp.status_code != 200:
        fail('error fetching tests_url content: {}'.format(resp.status_code))
    print('Got tests_url content: ', resp.text)

    # run tests_cmd, e.g. some script inside the downloaded content
    try:
        res['tests_cmd_out'] = str(subprocess.check_output(test_payload['tests_cmd'], shell=True))
    except Exception as e:
        fail('error running tests_cmd: {}'.format(e))

    return res


if __name__ == '__main__':
    try:
        res = fetch_and_run()
        ok(res)
    except Exception as e:
        fail('uncaught err in fetch_and_run(): {}'.format(e))
